/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                      MODULE ICMP Protocol Service                      **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.22                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/


#include "icmpservice.h"
#include "types.h"
#include "ipservice.h"
#include "frameutil.h"
#include "trace.h"
#include "uart0.h"


#define Prot_ICMP                1               // Internet Control Message Protocol
#define Prot_TCP                 6               // Transmission Control Protocol
#define Prot_UDP                 17              // User Datagram Protocol

#define ICMP_EchoReq             0x0800
#define ICMP_EchoReply           0x0000

#define MAX_ICMP_EchoDataSize    60

#define ICMP_TypeCodeOffset           EtherPackedHeaderRecTSize + IPHeaderSize + 0
#define ICMP_CheckSumOffset           EtherPackedHeaderRecTSize + IPHeaderSize + 2
#define ICMP_DataOffset               EtherPackedHeaderRecTSize + IPHeaderSize + 4

#define ICMPHeaderSize                4


/***********************************************************************************************
************************************************************************************************
**                                                                                            **
**                                          ICMP SERVICR                                      **
**                                                                                            **
************************************************************************************************
***********************************************************************************************/

static void SendICMPEcho ( PtrIPLayerInstanceRecT IPLayerInstance ,
                           IPHeaderRecT *         IPHeader ,
                           PtrOSFrameRecT         Frame )
{
  EMACInstanceRecT * InterfInstance ;
  EtherPackedHeaderRecT EtherHeader ;
  PtrOSFrameRecT OutFrame ;
  USHORT ICMPDataSize ;
  USHORT CheckSum ;
  /*-------------------------------------------------------------------------*/
  if ( IPHeader -> TotalLength > MAX_ICMP_EchoDataSize )
  {
    ICMPDataSize = MAX_ICMP_EchoDataSize - IPHeaderSize - ICMPHeaderSize ;
  } /* if ... */
  else
  {
    if ( IPHeader -> TotalLength > ( IPHeaderSize + ICMPHeaderSize ) )
      ICMPDataSize = IPHeader -> TotalLength - IPHeaderSize - ICMPHeaderSize ;
    else
      return ;
  } /* if ... else */ ;
  InterfInstance = IPLayerInstance -> InterfaceInstance ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  if ( OutFrame )
  {
    GetEthHeader ( & EtherHeader , Frame ) ;
    StartEthFrame ( OutFrame , EtherHeader . SourceMACAddress ,
                    InterfInstance -> InterfaceMACAddress , FRAME_IP ) ;
    CreateIPHeader ( OutFrame , Prot_ICMP , IPLayerInstance -> MyLocalIPAddress ,
                     IPHeader -> SourceIPAddress ) ;
    PutUSHORTSwapField ( OutFrame , IPTotalLengthOffset , ICMPDataSize + IPHeaderSize + ICMPHeaderSize ) ;
    CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPVersOffset ] , IPHeaderSize ) ;
    PutUSHORTNoSwapField ( OutFrame , IPHeaderChecksumOffset , CheckSum ) ;
    AddUSHORTSwapField ( OutFrame , ICMP_EchoReply ) ;
    AddUSHORTSwapField ( OutFrame , 0 ) ;   /* checksum */
    AddByteArrayField ( OutFrame , & Frame -> Data [ ICMP_DataOffset ] , ICMPDataSize ) ;
    CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPDataOffset ] , ICMPDataSize + ICMPHeaderSize ) ;
    PutUSHORTNoSwapField ( OutFrame , ICMP_CheckSumOffset , CheckSum ) ;
    InterfInstance -> InterfaceSendFrame ( OutFrame ) ;
    InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
  } /* if */ ;
} /* SendICMPEcho */

UCHAR ICMPMessage01 [ ]        = "\r\nPrzetwarzanie pakietu ICMP\r\nTYPE=" ;
UCHAR ICMPMessage02 [ ]        = " CODE=" ;
UCHAR ICMPMessage03 [ ]        = " HCRC=" ;
UCHAR ICMPMessage04 [ ]        = "\r\nIDENT=" ;
UCHAR ICMPMessage05 [ ]        = " SEQ=" ;
UCHAR ICMPMessage06 [ ]        = "\r\nOdebrano prosbe o echo\r\n" ;
UCHAR ICMPMessage07 [ ]        = "\r\nOdebrano odpowiedz na prosbe o echo\r\n" ;


void ProcessICMPPacked ( PtrIPLayerInstanceRecT IPLayerInstance ,
                         IPHeaderRecT *         IPHeader ,
                         PtrOSFrameRecT         Frame )
{
  USHORT TypeAndCode ;
  USHORT HCRC ;
  USHORT Ident ;
  USHORT Seq ;
  /*-------------------------------------------------------------------------*/
  TypeAndCode = GetUSHORTSwapField ( Frame , IPDataOffset ) ;
  HCRC = GetUSHORTSwapField ( Frame , IPDataOffset + 2 ) ;
  Ident = GetUSHORTSwapField ( Frame , IPDataOffset + 4 ) ;
  Seq = GetUSHORTSwapField ( Frame , IPDataOffset + 6 ) ;
  UART0SendString ( ICMPMessage01 ) ;
  TraceUCHAR ( ( UCHAR ) ( TypeAndCode >> 8 ) ) ;
  UART0SendString ( ICMPMessage02 ) ;
  TraceUCHAR ( ( UCHAR ) ( TypeAndCode & 0xFF ) ) ;
  UART0SendString ( ICMPMessage03 ) ;
  TraceUSHORT ( HCRC ) ;
  UART0SendString ( ICMPMessage04 ) ;
  TraceUSHORT ( Ident ) ;
  UART0SendString ( ICMPMessage05 ) ;
  TraceUSHORT ( Seq ) ;
  switch ( TypeAndCode )
  {
    case ICMP_EchoReq      :
      UART0SendString ( ICMPMessage06 ) ;
      SendICMPEcho ( IPLayerInstance , IPHeader , Frame ) ;
      break ;
    case ICMP_EchoReply    :
      UART0SendString ( ICMPMessage07 ) ;
      if ( IPLayerInstance -> ICMPEchoProcessService )
        IPLayerInstance -> ICMPEchoProcessService ( IPLayerInstance , IPHeader , Frame ) ;
      break ;
    default :
      ;
  } /* case */ ;
} /* ProcessICMPPacked */


static USHORT StrLen ( const UCHAR * Str )
{
  const UCHAR * SaveStr = Str ;
  /*-------------------------------------------------------------------------*/
  while ( * SaveStr ++ )
    ;
  return ( SaveStr - Str - 1 ) ;
} /* StrLen */


void SetICMPEchoService ( IPLayerInstanceRecT *    IPLayerInstance ,
                          ProcessProtPServiceProcT EchoService )
{
  /*-------------------------------------------------------------------------*/
  IPLayerInstance -> ICMPEchoProcessService = EchoService ;
} /* SetICMPEchoService */


void SendPingTo ( IPLayerInstanceRecT * IPLayerInstance ,
                  ULONG                 IPAddress ,
            const UCHAR *               Data )
{
  EMACInstanceRecT * InterfInstance ;
  PtrOSFrameRecT OutFrame ;
  ULONG TargetAddress ;
  USHORT ICMPDataSize ;
  USHORT CheckSum ;
  /*-------------------------------------------------------------------------*/
  if ( ( IPAddress & IPLayerInstance -> MyLocalNetMask ) ==
       ( IPLayerInstance -> MyLocalIPAddress & IPLayerInstance -> MyLocalNetMask ) )
  {
    TargetAddress = IPAddress ;
  } /* if ... */
  else
  {
    TargetAddress = IPLayerInstance-> MyDefaultGWIPAddress ;
    if ( ! TargetAddress )
    {
      return ;
    } /* if */ ;
  } /* if ... else */ ;
  ICMPDataSize = StrLen ( Data ) + ICMPHeaderSize ;
  InterfInstance = IPLayerInstance -> InterfaceInstance ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  StartEthFrame ( OutFrame , NIL , InterfInstance -> InterfaceMACAddress , FRAME_IP ) ;
  CreateIPHeader ( OutFrame , Prot_ICMP , IPLayerInstance -> MyLocalIPAddress , IPAddress ) ;
  PutUSHORTSwapField ( OutFrame , IPTotalLengthOffset , ICMPDataSize + IPHeaderSize + ICMPHeaderSize ) ;
  CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPVersOffset ] , IPHeaderSize ) ;
  PutUSHORTNoSwapField ( OutFrame , IPHeaderChecksumOffset , CheckSum ) ;
  AddUSHORTSwapField ( OutFrame , ICMP_EchoReq ) ;
  AddUSHORTSwapField ( OutFrame , 0 ) ;    /* checksum */
  AddUSHORTSwapField ( OutFrame , 12 ) ;   /* identifier */
  AddUSHORTSwapField ( OutFrame , 34 ) ;   /* squence number */
  AddByteArrayField ( OutFrame , ( void * ) Data , ICMPDataSize - ICMPHeaderSize ) ;
  CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPDataOffset ] , ICMPDataSize + ICMPHeaderSize ) ;
  PutUSHORTNoSwapField ( OutFrame , ICMP_CheckSumOffset , CheckSum ) ;
  IPLayerInstance -> SendIPPackedService ( IPLayerInstance , OutFrame , TargetAddress ) ;
  InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
} /* SendPingTo */



